import 'dart:async';
import 'dart:html' as html;
import 'dart:io';
import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import "dart:convert";

///## **BagelDB** is a content management system with rich API features. For more info, visit here bageldb.com

/// *BagelDB* class is the base class for any bagelDB request, it must be given the api [token] created at app.bageldb.com
class BagelDB {
  String token;

  BagelDB(this.token);

  /// return a *BagelDBRequest* object for a specific [collection]
  BagelDBRequest collection(String collection) {
    return BagelDBRequest(bagelDB: this, collection: collection);
  }

  BagelMetaRequest schema(String collection) {
    return BagelMetaRequest(instance: this, collection: collection);
  }
}

/// The base bagelDB response class, the response from bagelDB is inserted into the class
class BagelResponse {
  dynamic data;
  int? itemCount;
  int? statusCode;
  BagelResponse(
      {required this.data, required this.statusCode, this.itemCount = 0});

  @override
  String toString() {
    String str = 'status code: ${this.statusCode.toString()}'
        'ֿ\nitem count: ${this.itemCount.toString()}'
        '\ndata: ${this.data.toString()}';
    return str;
  }
}

class BagelMetaResponse {
  dynamic data;
  int? statusCode;

  BagelMetaResponse({required this.data, required this.statusCode});

  @override
  String toString() {
    String str = 'status code: ${this.statusCode.toString()}'
        '\ndata: ${this.data.toString()}';
    return str;
  }
}

class BagelMetaRequest {
  String collection;
  BagelDB instance;
  final String baseEndpoint = 'https://api.bagelstudio.co/api/public';
  Dio dio = Dio();

  BagelMetaRequest({required this.instance, required this.collection});

  Future<BagelMetaResponse> get() async {
    Options options = Options(headers: {
      'authorization': 'Bearer ${this.instance.token}',
      "Accept-Version": "v1"
    });
    String url = '$baseEndpoint/collection/${this.collection}/schema';
    Response response = await dio.get(url, options: options);
    BagelMetaResponse res = BagelMetaResponse(
      data: response.data,
      statusCode: response.statusCode,
    );
    return res;
  }
}

/// The builder for creating a bagelDB request, follow the builder pattern with the verbs get,
/// put, post and delete building and executing the request. There is also the ability to use the uploadImage builder,
/// for uploading images to an item

class BagelDBRequest {
  BagelDB bagelDB;
  String collection, sort, order;
  String itemID;
  int pageNum = 1, resPerPage = 100;
  List<String> queryLst = [];
  bool callEverything;
  final String baseEndpoint = 'https://api.bagelstudio.co/api/public';
  final String liveEndpoint = 'https://live.bageldb.com/api/public';
  Dio dio = Dio();

  BagelDBRequest(
      {required this.bagelDB,
      required this.collection,
      this.itemID = "",
      this.sort = "",
      this.order = "",
      this.resPerPage = 20,
      this.callEverything = false});

  /// Build and execute a get request to bagelDB, for both full collection and item requests
  Future<BagelResponse> get() async {
    Options options = Options();
    options.headers = {
      'authorization': 'Bearer ${this.bagelDB.token}',
      "Accept-Version": "v1"
    };

    Map<String, dynamic> params = Map<String, dynamic>();
    if (this.sort != "") params['sort'] = this.sort;
    if (this.order != "") params['order'] = this.order;
    if (this.callEverything)
      params['everything'] = this.callEverything.toString();
    params['pageNumber'] = this.pageNum.toString();
    params['perPage'] = this.resPerPage.toString();
    String item = this.itemID != "" ? '/' + this.itemID : "";
    String url = '$baseEndpoint/collection/${this.collection}/items$item';
    if (this.queryLst.isNotEmpty)
      url = url + "?query=" + this.queryLst.join("%2B");
    Response response =
        await dio.get(url, queryParameters: params, options: options);
    Headers headers = response.headers;
    BagelResponse res = BagelResponse(
      data: response.data,
      statusCode: response.statusCode!,
    );
    if (item == "") res.itemCount = int.tryParse(headers['item-count']![0]);
    return res;
  }

  /// Build and execute a post request to bagelDB. [item] should follow the collection schema as defined at app.bageldb.com
  Future<BagelResponse> post({@required Map<String, dynamic>? item}) async {
    var options = Options();
    options.headers = {
      'authorization': 'Bearer ${this.bagelDB.token}',
      "Accept-Version": "v1"
    };
    String url = '$baseEndpoint/collection/${this.collection}/items';
    return dio
        .post(url, data: item, options: options)
        .then((Response res) async {
      return BagelResponse(data: res.data, statusCode: res.statusCode!);
    });
  }

  /// Build and execute a image upload to bagelDB, for a specific item in a collection. [image] should be a valid file object and [slug] is the field to which the image will be updated. The item() methods must have been used in order to execute uploadImage
  Future<BagelResponse> uploadImage(String slug, File image) async {
    var options = Options();
    options.headers = {
      'authorization': 'Bearer ${this.bagelDB.token}',
      "Accept-Version": "v1"
    };

    String fileName = image.path.split('/').last;
    FormData formData = FormData.fromMap({
      "imageFile": await MultipartFile.fromFile(
        image.path,
        filename: fileName,
      ),
    });
    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}/image?imageSlug=$slug';
    Response res = await dio.put(url, data: formData, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode!);
  }

  /// Build and execute a put request to bagelDB. [item] should follow the collection schema as defined at app.bageldb.com
  Future<BagelResponse> put({@required Map<String, dynamic>? item}) async {
    var options = Options();
    options.headers = {
      'authorization': 'Bearer ${this.bagelDB.token}',
      "Accept-Version": "v1"
    };

    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}';
    Response res = await dio.put(url, data: item, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode!);
  }

  /// Build and execute a delete request to bagelDB.
  Future<BagelResponse> delete() async {
    var options = Options();
    options.headers = {
      'authorization': 'Bearer ${this.bagelDB.token}',
      "Accept-Version": "v1"
    };

    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}';
    Response res = await dio.delete(url, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode!);
  }

  /// quering the collection with a [key] and [value] parameters, by the [queryOperator]. The following are valid queryOperators '=', '!=', 'regex', '>', '<'. The value can also be a comma seperated list of values, allowing for an 'in' query.
  /// For example: ```query('age', '=', '32')``` or for an itemRef query ```query('position.itemRefID', '=', '[432EWDWE]')```
  query(String key, String queryOperator, dynamic value) {
    String query = key + ":" + queryOperator + ":" + value.toString();
    this.queryLst.add(Uri.encodeQueryComponent(query));
    return this;
  }

  /// sort data by [sortTerm] - a specific field, for example ```age```,
  /// and by [orderTerm] - ```desc``` for descending order, ```asc``` for ascending order.
  sortBy(String sortTerm, {String? orderTerm}) {
    this.sort = sortTerm;
    if (orderTerm != null) this.order = orderTerm;
    return this;
  }

  /// define the [resultsPerPage] - how many results will be in each page
  perPage(int resultsPerPage) {
    this.resPerPage = resultsPerPage;
    return this;
  }

  /// define the [pageNumber] - the index of the page
  pageNumber(int? pageNumber) {
    this.pageNum = pageNumber ?? 1;
    return this;
  }

  /// defines the request for a specific item by the item's [_id]
  item(String _id) {
    this.itemID = _id;
    return this;
  }

  /// specify whether the request should include all the details for item references
  everything() {
    this.callEverything = true;
    return this;
  }

  /// Live Data? No Problem

  StreamController<dynamic>? streamController;
  StreamSubscription? streamSubscription;

  StreamSubscription listen(Function(BagelEvent event) onData) {
    streamController = StreamController();
    Stream stream = streamController!.stream;
    _restartConnection();
    return stream.listen((event) => onData(event));
  }

  String requestID = "";
  _restartConnection() {
    streamSubscription?.cancel();

    String uri =
        '$liveEndpoint/collection/${this.collection}/live?authorization=${this.bagelDB.token}';
    uri = '$uri&requestID=$requestID';
    _listen(uri);
  }

  _listen(String uri) async {
    try {
      final eventSource = html.EventSource(uri, withCredentials: false);
      eventSource.onMessage.listen((html.MessageEvent event) =>
          streamController!.add(BagelEvent.fromPayload(event.data!)));
      eventSource.addEventListener("stop",
          (event) => streamController!.hasListener ? _restartConnection() : "");
      // eventSource.addEventListener("open",
      //     (event) => print((event as html.MessageEvent).data as String));
      // eventSource.onOpen.listen((html.Event event) {
        // requestID = (event as html.MessageEvent).data as String;
      // });
    } catch (err) {
      throw (err);
    }
  }
}

class BagelEvent {
  String? trigger, collectionID, organization, projectID, itemID;
  dynamic item;
  BagelEvent(
      {this.trigger,
      this.collectionID,
      this.organization,
      this.item,
      this.projectID,
      this.itemID});

  factory BagelEvent.fromPayload(String payload) {
    dynamic response = json.decode(payload);
    return BagelEvent(
        collectionID: response["collectionID"],
        projectID: response["projectID"],
        itemID: response["itemID"],
        trigger: response["trigger"],
        item: response["item"]);
  }
}
